#include <u.h>
#include <libc.h>
#include <avl.h>
#include <bio.h>
#include <regexp.h>
#include "utils.h"
#include "muffet.h"

void
usage(void)
{
	exits("usage");
}

	char* HTTPBASE = "http://www.cybersavior.dev/";
	char* GEMBASE  = "gemini://cybersavior.dev/";
	char* AUTHOR = "Takarabune";
	char* TITLE = "cybersavior(7)";
	
char * irefrx ="([^ ^	]*) *(.*)";
char * erefrx ="[#^]([^	 ]*) *([^ 	]*) *(.*)";
char * urlrx ="[hH][tT][Tt][pP][sS]?://([!-~]+) *(.*)?";
char * namerx = "^([^	 ].*)";
char * eofrx = "^[ 	]*!";
char * valrx = "^[ 	](.*)";
char * listrx ="^[ 	][ 	](.*)";
char * linkrx ="^[ 	]?[	 ]([^:]*) : [hH][Tt]+[pP][sS]?://([!-~]+) *(.*)?";

Reprog * irefrp;
Reprog * erefrp;
Reprog * urlrp;
Reprog * namerp;
Reprog * eofrp;
Reprog * valrp;
Reprog * listrp;
Reprog * linkrp;

Avltree *tree;

char * muffet_type_str[]= {"hub", "article", "text", "tag", "portfolio", "album", "literal", "feed", "other", "ERROR"};

muffet_type
getmuffettype(char * s){
		for(int i = 0; i < (int)o_ERROR; i++)
			if(mcuistrcmp(s, muffet_type_str[i]) == 0)
				return (muffet_type)i;
	return o_ERROR;
}

char * muffet_reftype_str[]={"intern", "img", "snd", "vid", "file", "txt", "html", "list", "archive", "yt", "http", "bandcamp", "proto", "ERROR"};

muffet_reftype
getreftype(char * s){
		for(int i = 0; i < (int)r_ERROR; i++)
			if(mcuistrcmp(s, muffet_reftype_str[i]) == 0)
				return (muffet_reftype)i;
	return r_ERROR;
}

void
printmuffetinc(Avl *n, Biobuf *, int out, int){
	muffet_inc_ref * iR = (muffet_inc_ref*) n;
	fprint(out, "\t\t%s\n", iR->name);
}

void
printmuffetref(Avl *n, Biobuf *, int out, int){
	muffet_out_ref * oR = (muffet_out_ref*) n;
	if(oR->desc){
		fprint(out, "\t\t[%d]%s - %s\n", oR->index, oR->desc, oR->name);
	} else {
		fprint(out, "\t\t[%d]%s\n", oR->index, oR->name);
	}
}

void
printmuffettag(Avl *n, Biobuf *, int out, int){
	muffettag *T = (muffettag*) n;
	fprint(out, "\t\t%s\n", T->name);
}

void
printmuffetlink(Avl *n, Biobuf *, int out, int){
	muffetlink *L = (muffetlink*) n;
	fprint(out, "\t\t%s : %s\n", L->name, L->url);
}

void
printmuffetchild(Avl *n, Biobuf *, int out, int){
	muffetchild *C = (muffetchild*) n;
	fprint(out, "\t\t%s\n", C->name);
}

void
printmuffetbody(Avl *n, Biobuf *, int out, int){
	muffetbody *C = (muffetbody*) n;
	if(C->refcount == 0){
		fprint(out, "\t\t%s\n", C->line);
	} else {
		fprint(out, "\t\t");
		for(int i = 0; i < C->refcount; ++i){
			fprint(out, "%s{%s}", C->parts[i],C->refs[i]);
		}
		fprint(out, "%s\n",C->parts[C->refcount]);
	}
}

void
printmuffetnode(Avl *n, Biobuf *in, int out, int dir){
	muffetnode * N = (muffetnode*) n;
	int fd = 0;
	if(N->host){
		if(N->type == o_TEXT){
			 fd = opentext(N->name);
		}
		fprint(out, "\n%s\n", N->name);
		fprint(out, "\tHOST : %s\n", N->host);
		fprint(out, "\tBREF : %s\n", N->bref);
		fprint(out, "\tTYPE : %s\n", muffet_type_str[(int)N->type]);
		if(avlmax(N->children)){
			fprint(out, "\tCHILDREN\n");
			walk(N->children, in, out, printmuffetchild, dir);
		}
		if(avlmax(N->bodylines)){
			fprint(out, "\tBODY\n");
			walk(N->bodylines, in, out, printmuffetbody, 0);
		}
		if(avlmax(N->links)){
			fprint(out, "\tLINK\n");
			walk(N->links, in, out, printmuffetlink, 0);
		}
		if(avlmin(N->tags)){
			fprint(out, "\tTAGS\n");
			walk(N->tags, in, out, printmuffettag, dir);
		}
		if(avlmin(N->inc_refs)){
			fprint(out, "\tINC\n");
			walk(N->inc_refs, in, out, printmuffetinc, dir);
		}
		if(avlmin(N->out_refs)){
			fprint(out, "\tREF\n");
			walk(N->out_refs, in, out, printmuffetref, dir);
		}
		if(fd>0){
			Biobuf * txt = Bfdopen(fd, OREAD);
			char * line;
			while(line = Brdstr(txt, '\n', 0)){
				fprint(out, "%s", line);
			}
			fprint(out, ".\n");
		}
	} else {
		fprint(out, "\n%s\n", N->name);
		if(avlmax(N->links)){
			fprint(out, "\tLINK\n");
			walk(N->links, in, out, printmuffetlink, 0);
		}
	}
}

void
writeFeed(Biobuf *in, int out){
	muffetnode key = {.name = "FEED"};
	muffetnode* F = (muffetnode*)avllookup(tree, &key, 0);
	char * linepage;
	char * linepub;
	char * linedate;
	if (F) {
		int fd_rss  = openlinks("rss.xml");
		int fd_atom = openlinks("atom.xml");
		int fd_gmi  = opengem("feed");
		int fd_html = openhtml("feed");
		
		fprint(fd_atom, "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
		fprint(fd_rss , "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
		fprint(fd_atom, "<feed xmlns=\"http://www.w3.org/2005/Atom\">\n\n");
		fprint(fd_rss , "<rss version=\"2.0\">\n\n");
		fprint(fd_atom, "\t<title>%s</title>\n", TITLE);
		fprint(fd_rss , "\t<channel>\n\t<title>%s</title>\n", TITLE);
		fprint(fd_gmi , "#%s\n\n", TITLE);
		fprint(fd_html, "<!DOCTYPE html> \n<html lang='en'>\n<head>\n");
		fprint(fd_html, "\t<link rel='stylesheet' type='text/css' href='../links/main.css'>\n");
		fprint(fd_html, "\t<link rel='icon' type='image/x-icon' href='../media/identity/favicon.ico'>\n");
		fprint(fd_html, "\t<title>%s - Feed</title>\n", TITLE);
		if(bgexists("feed") > 0)
			fprint(fd_html, "\t<style>body{background-image: url('../media/bg/feed.png'); z-index: -10; -o-background-size: cover; background-size: cover;}</style>\n");
		{
			fprint(fd_html, "</head>\n\n<div class='menu'>\n\t<header><a href='home.html'><img src='../media/identity/banner.png' alt='%s' height='50'></a>\n\t<nav>\n", TITLE);
			muffetnode key = {.name="feed"};
			muffetnode *F, *H, *G;
			char * cannon;

			F = (muffetnode*)avllookup(tree, &key, 0);
				
			if(F && F->host){
				key.name = F->host;
				H = (muffetnode*)avllookup(tree, &key, 0);
				if(H){
					key.name = H->host;
					if(H->host)
						G = (muffetnode*)avllookup(tree, &key, 0);
					else 
						G = nil;
					if(G){
						muffetchild* C = (muffetchild*)avlmin(G->children);
						fprint(fd_html, "\t\t<ul class='up'>\n");
						do{
							cannon = cannonizename(C->name);
							int b = mcuistrcmp(C->name, H->name) == 0;
							fprint(fd_html, "\t\t\t<il>%s<a href='%s.html'>%s</a>%s</il>\n",
								b ? "<b>" : "", cannon, C->name, b ? "</b>" : "");
							free(cannon);
						}while(C = (muffetchild*)avlnext(C));
						fprint(fd_html, "\t\t</ul>\n\&emsp;");
					}else{
						cannon = cannonizename(H->name);
						fprint(fd_html, "\t\t<ul class='up'>\n\t\t\t<il><b><a href='%s.html'>%s</a></b></il>\n\t\t</ul>\n\&emsp;", cannon, H->name);
						free(cannon);
					}
					fprint(fd_html, "\n\t\t<ul class='sib'>\n");
					muffetchild* C = (muffetchild*)avlmin(H->children);
					//we know C has to exist, since node is one of them
					do{
						cannon = cannonizename(C->name);
						int b = mcuistrcmp(C->name, F->name) == 0;
						fprint(fd_html, "\t\t\t<il>%s<a href='%s.html'>%s</a>%s</il>\n",
								b ? "<b>" : "", cannon, C->name, b ? "</b>" : "");
						free(cannon);
					} while(C = (muffetchild*)avlnext(C));
					fprint(fd_html, "\t\t</ul>\n");
				}
			}
			fprint(fd_html, "\t</nav>\n</header>\n</div></br>\n<body class='main'>\n<div>\n\n<h1>FEED</h1>\n");

		}

		

		fprint(fd_atom, "\t<link href=\"%s\"/>\n", GEMBASE);
		fprint(fd_rss , "\t<link href=\"%s\"/>\n", HTTPBASE);

		fprint(out, "l\n%s\n\tDATE\n!\n", ((muffetchild*)avlmax(F->children))->node->name);
		linedate = Brdstr(in, '\n', 1);
		fprint(out, "l\n%s\n\tPUB\n!\n", ((muffetchild*)avlmax(F->children))->node->name);
		linepub = Brdstr(in, '\n', 1);
		match[0].sp = match[0].ep = 0; regexec(valrp, linedate, match, MSZ);
		linedate = match[1].sp;
		match[0].sp = match[0].ep = 0; regexec(valrp, linepub, match, MSZ);
		linepub = match[1].sp;

		fprint(fd_atom, "\t<updated>%sT12:00:00Z</updated>\n", linedate);
		fprint(fd_rss, "\t<lastBuildDate>%s 00:00:00 +0800</lastBuildDate>\n", linepub);
		free(linedate);
		free(linepub);
	
		fprint(fd_atom, "\t<id>\"%s\"<\id>\n\n", GEMBASE);
		fprint(fd_rss , "\t<image> <url>%smedia/services/rss.png<\url> <\image>\n", HTTPBASE);

		muffetchild* C = (muffetchild*)avlmax(F->children);
		do{
			muffetnode* c = C->node;
			fprint(fd_atom, "\t<entry>\n");
			fprint(fd_rss , "\t<item>\n");
			fprint(fd_atom, "\t\t<title>%s</title>\n", c->bref);
			fprint(fd_atom, "\t\t<author><name>%s</name></author>\n", AUTHOR);
			fprint(fd_rss , "\t\t<title>%s</title>\n", c->bref);
			fprint(fd_rss , "\t\t<author>%s</author>\n", AUTHOR);
		
			fprint(out, "l\n%s\n\tPAGE\n!\n", c->name);

			linepage = Brdstr(in, '\n', 1);
			match[0].sp = match[0].ep = 0; regexec(valrp, linepage, match, MSZ);
			linepage = match[1].sp;

			fprint(fd_atom, "\t\t<link rel=\"alternate\" href=%s%s.gmi />\n", GEMBASE, linepage);
			fprint(fd_atom, "\t\t<id>\"%s%s.gmi\"</id>\n", GEMBASE, linepage);
			fprint(fd_rss , "\t\t<link>%ssite/%s.html</link>\n", HTTPBASE, linepage);
			fprint(fd_rss , "\t\t<guid isPermaLink=\"false\">%ssite/%s.html</guid>\n", HTTPBASE, linepage);
		
			fprint(out, "l\n%s\n\tDATE\n!\n", c->name);
			linedate = Brdstr(in, '\n', 1);
			match[0].sp = match[0].ep = 0; regexec(valrp, linedate, match, MSZ);
			linedate = match[0].sp;
			fprint(out, "l\n%s\n\tPUB\n!\n", c->name);
			linepub = Brdstr(in, '\n', 1);
			match[0].sp = match[0].ep = 0; regexec(valrp, linepub, match, MSZ);
			linepub = match[1].sp;

			fprint(fd_atom, "\t\t<updated>%sT12:00:00Z</updated>\n", linedate);
			fprint(fd_rss , "\t\t<pubDate>%s 00:00:00 +0800</pubdate>\n", linepub);
			fprint(fd_gmi, "=>%s%s.gmi %s - %s\n", GEMBASE, linepage, linedate, c->bref);
			fprint(fd_html, "<p><a href='./%s.html'>%s - %s</a></p>\n", linepage, linedate, c->bref);
			free(linepage);
			free(linedate);
			muffetbody * body = (muffetbody*) avlmin(c->bodylines);
			if(body){
				fprint(fd_atom, "\t\t<summary>\n");
				fprint(fd_rss, "\t\t<description>\n");
					do{
					printmuffetbody(body, nil, fd_atom, 0);
					printmuffetbody(body, nil, fd_rss, 0);
				} while(body = (muffetbody*)avlnext(body));
				fprint(fd_atom, "\t\t</summary>\n");
				fprint(fd_rss, "\t\t</description>\n");
			}

			fprint(fd_atom, "\t</entry>\n\n");
			fprint(fd_rss, "\t</item>\n\n");
		}while(C = (muffetchild*)avlprev(C));
		fprint(fd_atom, "</feed>\n\n");
		fprint(fd_rss , "\t</channel>\n</rss>\n");
		fprint(fd_html, "\n<h3>Historical records</h3>\n<p>RSS feed <a href='%slinks/rss.xml' title='rss.xml'>[rss.xml]</a></p>\n</br>\n</div>\n</body>", HTTPBASE);
		fprint(fd_html, "\n<footer>\n<div><a href='./license.html' title='copydown'>&Omega;</a>&emsp;&emsp;&emsp;&emsp;Fragment of <a href='./denpa_webring.html' title='Denpa Webring'>The Denpa Webring</a>&emsp;&emsp;&emsp;&emsp;<a href='./myugii_fannan.html'><img src='../media/img/services/myugii.gif' alt='myugii.cybersavior.dev' height='31' width='88'></a></div>\n</footer>\n</html>\n");
	
		close(fd_atom);
		close(fd_rss);
		close(fd_gmi);
		close(fd_html);
	}
}

void
doHTML(void){
	muffetnode* N = (muffetnode*)avlmin(tree);
	do{
		if(N->type == o_ERROR )
			continue;
		int fd = openhtml(N->name);
		char * cannon;
		
		fprint(fd, "<!DOCTYPE html> \n<html lang='en'>\n<head>\n");
//			fprint(fd, "<meta >");
		fprint(fd, "\t<link rel='stylesheet' type='text/css' href='../links/main.css'>\n");
		fprint(fd, "\t<link rel='icon' type='image/x-icon' href='../media/identity/favicon.ico'>\n");
		fprint(fd, "\t<title>%s - %s</title>\n", TITLE, N->name);
			
		cannon = cannonizename(N->name);
		if(bgexists(cannon) > 0) 
			fprint(fd, "\t<style>body{background-image: url('../media/bg/%s.png'); z-index: -10; -o-background-size: cover; background-size: cover;}</style>\n", cannon);
		free(cannon);
		fprint(fd, "</head>\n\n");
		fprint(fd, "<div class='menu'>\n");
		fprint(fd, "\t<header><a href='home.html'><img src='../media/identity/banner.png' alt='%s' height='50'></a>\n", TITLE);
			
		muffetnode key;
		muffetnode* F;

		//parent and it's siblings
		if(N->host){
			fprint(fd, "\t<nav>\n\t\t<ul class='up'>\n");
				//find host's host.
		
			key.name = N->host;
			F = (muffetnode*)avllookup(tree, &key, 0);
			if(F && F->host){
				key.name = F->host;
				F = (muffetnode*)avllookup(tree, &key, 0);
				if(F){ 
					muffetchild* C = (muffetchild*)avlmin(F->children);
					do{
						cannon = cannonizename(C->name);
						int b = mcuistrcmp(C->name, N->host) == 0;
						fprint(fd, "\t\t\t<il>%s<a href='%s.html'>%s</a>%s</il>\n",
							b?"<b>":"", cannon, C->name, b?"</b>":"");
						free(cannon);
					}while(C = (muffetchild*)avlnext(C));
				
				}
			}else if(F){
				cannon = cannonizename(F->name);
				fprint(fd, "\t\t\t<il><b><a href='%s.html'>%s</a></b></il>\n",
						cannon, F->name);
				free(cannon);
			}
			fprint(fd, "\t\t</ul>\n\&emsp;");
		}
			//siblings and item
		if(N->host){
			fprint(fd, "\n\t\t<ul class='sib'>\n");
				//find host.
			key.name = N->host;
			F = (muffetnode*)avllookup(tree, &key, 0);
			if(F){
				muffetchild* C = (muffetchild*)avlmin(F->children);
				//we know C has to exist, since node is one of them
				do{
					cannon = cannonizename(C->name);
					int b = mcuistrcmp(C->name, N->name) == 0;
				 	fprint(fd, "\t\t\t<il>%s<a href='%s.html'>%s</a>%s</il>\n",
							b ? "<b>" : "", cannon, C->name, b ? "</b>" : "");
					free(cannon);
				}while(C = (muffetchild*)avlnext(C));
			}
			fprint(fd, "\t\t</ul>\n\&emsp;");
		}
		//children
		fprint(fd, "\n\t\t<ul class='child'>\n");
		muffetchild* C = (muffetchild*)avlmin(N->children);
		if(C){
			do{
				cannon = cannonizename(C->name);
				fprint(fd, "\t\t\t<il><a href='%s.html'>%s</a></il>\n",
						cannon, C->name);
				free(cannon);		
			}while(C = (muffetchild*)avlnext(C));
		}
		fprint(fd, "\t\t</ul>\n");
		fprint(fd, "\t</nav>\n</header>\n</div>\n");

		//body
		fprint(fd, "</br><body class='main'>\n<div>\n\n\n<h1>%s</h1>\n", N->name);
		if(N->bref){
			fprint(fd, "\n\n\n<h3>%s</h3>\n",N->bref);
		}
		muffetbody* B = (muffetbody*)avlmin(N->bodylines);
		if(B){
			int s = 1;
			fprint(fd, "\n");
			do{
				fprint(fd, "\n\t<p>");
				int i = 0;
				do{
					fprint(fd, "%s", B->parts[i]);
					if(i >= (B->refcount)){
						i = i+1;
						continue;
					}
					muffet_out_ref  k = {.index=s};
					muffet_out_ref* r = (muffet_out_ref*)avllookup(N->out_refs, &k, 0);
					if(!r){
						i = i+1;
						continue;
					}
					cannon = cannonizename(r->name);
					switch(r->type){
						case r_INTERN:
		fprint(fd, "<a href='%s.html' title='%s'>[%s]</a>",
				cannon, r->desc?r->desc:r->name, r->name);
							break;
						case r_IMG:
		fprint(fd, "<img max-height='500' src='%smedia/img/%s' alt='%s'>",
				HTTPBASE, cannon, r->desc?r->desc:r->name);
							break;
						case r_SOUND:
		fprint(fd, "<audio controls><source src='%smedia/sound/%s.mp3' type='audio/mpeg'><source src='%smedia/sound/%s.ogg' type='audio/ogg'><a href='%smedia/sound/%s.mp3'>MP3</a></audio>", 
				HTTPBASE, cannon, HTTPBASE, cannon, HTTPBASE, cannon);
							break;
						case r_VIDEO:
		fprint(fd, "<video height='600' controls><source src='%smedia/video/%s.mp4 type='video/mp4'><source src='%smedia/video/%s.mov' type='video/mp4'><a href='%smedia/video/%s.mov>[MOV]</a> <a href='%smedia/video/%s.mp4'>[mp4]</a></video>",
				HTTPBASE, cannon, HTTPBASE, cannon, HTTPBASE, cannon, HTTPBASE, cannon);
							break;
						case r_FILE:
				fprint(fd, "<a href='%s%s' title='%s'>[%s]</a>",
						HTTPBASE, cannon, r->desc?r->desc:r->name, r->name);
							break;
						case r_TXT:
				fprint(fd, "<a href='%smedia/txt/%s' title='%s'>[%s]</a>",
						HTTPBASE, cannon, r->desc?r->desc:r->name, r->name);
							break;
						case r_HTML:
				fprint(fd, "<a href='%smedia/html/%s' title='%s'>[%s]</a>",
						HTTPBASE, cannon, r->desc?r->desc:r->name, r->name);
							break;
						case r_LIST:
				fprint(fd, "<a href='%s.html' title='%s'>[%s]</a>",
						r->name, r->desc?r->desc:r->name, r->name);
							break;
						case r_ARCHIVE:
		fprint(fd, "<iframe src='https://archive.org/embed/%s' width='640' height='480' frameborder='0' allowfullscreen> %s - %s </iframe>",
				r->name, r->desc?r->desc:r->name, r->name);
							break;
						case r_YOUTUBE:
		fprint(fd, "<a href='http://www.youtube.com/watch?v=%s' title='%s'>[%s]</a>",
				r->name, r->desc?r->desc:r->name, r->name);
							break;
						case r_HTTP:
		fprint(fd, "<a href='http://%s' title='%s'>[%s]</a>",
				r->name, r->desc?r->desc:r->name, r->name);
							break;
						case r_BANDCAMP:
		fprint(fd, "<iframe style='height:50px;' src='https://bandcamp.com/EmbeddedPlayer/album=%s' seamless> %s - %s </iframe>",
				r->name, r->desc?r->desc:r->name, r->name);
							break;
						case r_PROTOCOL:
						case r_ERROR:
						default:
							break;
					}
					free(cannon);
					s = s+1;
					i = i+1;
				}while(i <= B->refcount);
				fprint(fd, "</p>\n");
			}while(B = (muffetbody*)avlnext(B));
			fprint(fd, "</br>\n");
		}
		if(N->type == o_TEXT){
			fprint(fd, "\n<p style='white-space: pre-wrap'>");
			int tfd = opentext(N->name);
			if(tfd>0){
				Biobuf * txt = Bfdopen(tfd, OREAD);
				char * line;
				while(line = Brdstr(txt, '\n', 0)){
					fprint(fd, "%s", line);
				}
				fprint(fd, ".\n");
			}
			close(tfd);
			fprint(fd, "</p>");
		}
			//links
		muffetlink* L = (muffetlink*)avlmin(N->links);
		if(L){
			fprint(fd, "\n\t<h3>additional links</h3>\n");
			do{
				fprint(fd, "<a href='http://%s' title='%s'>%s</a></br>\n",
					L->url, L->desc?L->desc:"", L->name );
			}while(L = (muffetlink*)avlnext(L));
			fprint(fd, "<br>\n");
		}
			//refs
		muffet_inc_ref* iR = (muffet_inc_ref*)avlmin(N->inc_refs);
		if(iR){
			fprint(fd, "\n\t<h3>incoming references</h3>\n");
			do{
				cannon = cannonizename(iR->name);
				fprint(fd, "\t\t<a href='%s.html' title='%s'>%s</a></br>\n",
		cannon, iR->node?(iR->node->bref?iR->node->bref:""):"", iR->name);
				free(cannon);
			}while(iR = (muffet_inc_ref*)avlnext(iR));
			fprint(fd, "<br>\n");
		}
			//tags
		muffettag* T = (muffettag*)avlmin(N->tags);
		if(T){
			fprint(fd,"\t\n<h2>tags</h2>\n");
			do{
				cannon = cannonizename(T->name);
				fprint(fd, "\t\t<a href='%s.html' title='%s'>%s</a></br>\n",
						cannon, T->node?T->node->bref:" ", T->name );
				free(cannon);
			}while(T = (muffettag*)avlnext(T));
		}
		fprint(fd, "\n</div>\n</body>");
		fprint(fd, "\n<footer>\n<div><a href='./license.html' title='copydown'>&Omega;</a>&emsp;&emsp;&emsp;&emsp;Fragment of <a href='./denpa_webring.html' title='Denpa Webring'>The Denpa Webring</a>&emsp;&emsp;&emsp;&emsp;<a href='./myugii_fannan.html'><img src='../media/img/services/myugii.gif' alt='myugii.cybersavior.dev' height='31' width='88'></a></div>\n</footer>\n</html>\n");
		close(fd);
	}while(N = (muffetnode*)avlnext(N));
}


void
doGemini(void){
	
	muffetnode* N = (muffetnode*)avlmin(tree);
	do{	
		if(N->type == o_ERROR )
			continue;
		int fd = opengem(N->name);
		char * cannon;
		if(N->host){
			cannon = cannonizename(N->host);
			fprint(fd, "###host\n");
			fprint(fd, "=>%s%s.gmi %s\n", GEMBASE, cannon, N->host);
			free(cannon);
			cannonizename(N->name);
			if(bgexists(cannon) > 0) 
				fprint(fd, "=>%smedia/bg/%s.png background\n", GEMBASE, cannon);
			free(cannon);
		}
		muffetchild* C = (muffetchild*)avlmin(N->children);
		if(C){
			fprint(fd, "###children\n");
			do{
				if(C->node->type <= o_OTHER){
					cannon = cannonizename(C->name);
					fprint(fd, "=>%s%s.gmi %s - %s\n",
							GEMBASE, cannon, C->name, C->node->bref);
					free(cannon);
				}
			}while(C = (muffetchild*)avlnext(C));
		}

		muffettag* T = (muffettag*)avlmin(N->tags);
		if(T){
			fprint(fd,"\n###tags\n");
			do{
				cannon = cannonizename(T->name);
				fprint(fd, "=>%s%s.gmi %s - %s\n",
						GEMBASE, cannon, T->name, T->node?T->node->bref:" " );
				free(cannon);
			}while(T = (muffettag*)avlnext(T));
		}


		fprint(fd, "\n\n\n#%s\n", N->name);
		if(N->bref) fprint(fd, "##%s\n", N->bref);

		muffetbody* B = (muffetbody*)avlmin(N->bodylines);
		if(B){
			int s = 1;
			int e = 1;
			fprint(fd, "\n");
			do{
				e = e+B->refcount;
				for(int i = 0; i < B->refcount; ++i){
					//fprint(2, "%s[%s]", B->parts[i], B->refs[i]);
					fprint(fd, "%s[%s]", B->parts[i], B->refs[i]);
				}
				fprint(fd,"%s\n", B->parts[B->refcount]);
				for(; s<e; ++s){
					muffet_out_ref k = {.index=s};
					muffet_out_ref* r = (muffet_out_ref*)avllookup(N->out_refs, &k, 0);
					cannon = cannonizename(r->name);
					switch(r->type){
						case r_INTERN:
			fprint(fd, "=>%s%s.gmi [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break;
						case r_IMG:
						case r_SOUND:
						case r_VIDEO:
			fprint(fd, "=>%smedia/%s [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break;
						case r_FILE:
			fprint(fd, "=>%s%s [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break;
						case r_TXT:
			fprint(fd, "=>%smedia/txt/%s [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break;
						case r_HTML:
			fprint(fd, "=>%smedia/html/%s [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break;
						case r_LIST:
			fprint(fd, "=>%s%s.gmi [%d] %s\n",
								GEMBASE, cannon, s, r->desc?r->desc:r->name);
									break; //this isn't right, but should work for now.
						case r_ARCHIVE:
			fprint(fd, "=>http://archive.org/details/%s [%d] %s\n",
								r->name, s,  r->desc?r->desc:"");
									break;
						case r_YOUTUBE:
			fprint(fd, "=>http://www.youtube.com/watch?v=%s [%d] %s\n",
								r->name, s,  r->desc?r->desc:"");
									break;
						case r_HTTP:
			fprint(fd, "=>http://%s [%d] %s\n", r->name, s,  r->desc?r->desc:"");
									break;
						case r_BANDCAMP:
			fprint(fd, "=>https://bandcamp.com/EmbeddedPlayer/album=%s [%d] %s",
								r->name, s,  r->desc?r->desc:"");
						case r_PROTOCOL:
						case r_ERROR:
						default:
									break;
					}

					free(cannon);
				}
			}while(B = (muffetbody*)avlnext(B));
		}
		if(N->type == o_TEXT){
			int tfd = opentext(N->name);
			if(tfd>0){
				Biobuf * txt = Bfdopen(tfd, OREAD);
				char * line;
				while(line = Brdstr(txt, '\n', 0)){
					fprint(fd, "%s", line);
				}
				fprint(fd, ".\n");
			}
			close(tfd);
		}


		muffetlink* L = (muffetlink*)avlmin(N->links);
		if(L){
			fprint(fd, "\n###additional links\n");
			do{
				fprint(fd, "=>http://%s %s - %s\n",
						L->url, L->name, L->desc?L->desc:"" );
			}while(L = (muffetlink*)avlnext(L));
		}
		
		muffet_inc_ref* iR = (muffet_inc_ref*)avlmin(N->inc_refs);
		if(iR){
			fprint(fd, "\n###incoming references\n");
			do{
				cannon = cannonizename(iR->name);
				fprint(fd, "=>%s%s.gmi %s - %s\n", GEMBASE, cannon, iR->name, iR->node?(iR->node->bref?iR->node->bref:""):"");
				free(cannon);
			}while(iR = (muffet_inc_ref*)avlnext(iR));
		}
		
		close(fd);
		
	}while(N = (muffetnode*)avlnext(N));
}

void
freemuffetnode(Avl * n){
	muffetnode		* N = (muffetnode *) n;
	muffetchild		* C;
	muffetbody		* B;
	muffetlink		* L;
	muffet_inc_ref	* iR;
	muffet_out_ref	* oR;
	muffettag		* T;
	
	muffetchild Ckey	 =	{ .name = ""};
	muffetbody Bkey		 =	{ .index = 0};
	muffetlink Lkey 	 =	{ .index = 0};
	muffet_inc_ref iRkey =	{ .name = ""};
	muffet_out_ref oRkey =	{ .index = 0};
	muffettag Tkey		 =	{ .name = ""};

	free(N->bref);
	free(N->host);
	free(N->name);

	while(C = (muffetchild*) avllookup(N->children, &Ckey, -1)){
		C = (muffetchild*) avldelete(N->children, C);
		freemuffetchild(C);
	}
	while(B = (muffetbody*) avllookup(N->bodylines, &Bkey, 1)){
		B = (muffetbody*) avldelete(N->bodylines, B);
		freemuffetbody(B);
	}
	while(L = (muffetlink*) avllookup(N->links, &Lkey, 1)){
		L = (muffetlink*) avldelete(N->links, L);
		freemuffetlink(L);
	}
	while(iR = (muffet_inc_ref*) avllookup(N->inc_refs, &iRkey, -1)){
		iR = (muffet_inc_ref*) avldelete(N->inc_refs, iR);
		freemuffet_inc_ref(iR);
	}
	while(oR = (muffet_out_ref*) avllookup(N->out_refs, &oRkey, 1)){
		oR = (muffet_out_ref*) avldelete(N->out_refs, oR);
		freemuffetchild(oR);
	}
	while(T = (muffettag*) avllookup(N->tags, &Tkey, -1)){
		T = (muffettag*) avldelete(N->tags, T);
		freemuffettag(T);
	}

	free(N->children);
	free(N->bodylines);
	free(N->links);
	free(N->inc_refs);
	free(N->out_refs);
	free(N->tags);
	free(N);
}
void
freemuffetchild (Avl * n){
	muffetchild	*N = (muffetchild *) n;
	//free(N->name); these belong to the child
	//free(N->node);
	free(N);
}
void
freemuffetbody (Avl * n){
	muffetbody	*N = (muffetbody *) n;
	free(N->line);
	free(N->refs);
	free(N->parts);
	free(N);
}
void
freemuffetlink (Avl * n){
	muffetlink	*N = (muffetlink *) n;
	free(N->name);
	free(N->url);
	free(N);
}
void
freemuffet_inc_ref (Avl * n){
	muffet_inc_ref	*N = (muffet_inc_ref *) n;
	//free(N->name); belongs to the node
	//don't free the node
	free(N);
}
void
freemuffet_out_ref (Avl * n){
	muffet_inc_ref	*N = (muffet_inc_ref *) n;
	//free(N->name);
	//free(N->desc); //these belong to body's line
	free(N);
}
void
freemuffettag (Avl * n){
	muffet_inc_ref	*N = (muffet_inc_ref *) n;
	free(N->name);
	free(N);
}




void
populateNAME(Biobuf * in, int out){
	muffetnode * recentmn;
	Avl * dup;

	char * line;

	fprint(out, "l\n!\n");
	while( line = Brdstr(in, '\n', 1)){
		match[0].sp = match[0].ep = 0;
		if(regexec(eofrp, line, match, MSZ)){
		//	print("got to end\n");
			free(line);
			break;
		}
	
		match[0].sp = match[0].ep = 0;
		if(regexec(namerp, line, match, MSZ)){
			recentmn = (muffetnode *)malloc(sizeof(muffetnode));
			recentmn->name = copymatch(1);
			recentmn->host = nil;
			recentmn->bref = nil;
			recentmn->type = o_OTHER;
			recentmn->children = avlcreate(mccmp);
			recentmn->bodylines = avlcreate(mbcmp);
			recentmn->links = avlcreate(mlcmp);
			recentmn->inc_refs = avlcreate(mircmp);
			recentmn->out_refs = avlcreate(morcmp);
			recentmn->tags = avlcreate(mtcmp);
			if(dup = avlinsert(tree, recentmn)){
				freemuffetnode(dup);
			}
		}
		free(line);
	}
}

void
populateHOST(Avl * n, Biobuf *in, int out, int){
	muffetnode * N = (muffetnode *) n;
	muffetnode * H;
	muffetnode Hkey;
	muffetchild * NaC;

	char * line;
	char* hostlustr = "\nl\n%s\n\tHOST\n!\n";

	fprint(out, hostlustr, N->name);
	line = Brdstr(in, '\n', 1);
	match[0].sp = match[0].ep = 0;
	if(regexec(valrp, line, match, MSZ)){
		N->host = copymatch(1);
		Hkey.name = N->host;
		H = (muffetnode*)avllookup(tree, &Hkey, 0);
		if(H){
			NaC = (muffetchild *)malloc(sizeof(muffetchild));
			NaC->name = N->name;
			NaC->node = N;
			NaC = (muffetchild *)avlinsert(H->children, NaC);
			if(NaC)
				freemuffetchild(NaC);
		}
		free(line);
	} else { // if it's not returned on the first line then we need to clear till eof since it means the node has no host kvpair and the whole node was just returned to us.
		while( line = Brdstr(in, '\n', 1)){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				free(line);
				break;
			}
			free(line);
		}
	}
}

void indexPrint(muffetnode* N , int out, int depth){
	char *cannon = cannonizename(N->name);
	fprint(out, "		{%s |", cannon);
	free(cannon);
	for(int i = 1; i < depth; ++i) fprint(out, "==");	
	fprint(out, "%s}\n", N->name);
	muffetchild * current = (muffetchild*)avlmin(N->children);
	if(current && current->node != N){
		do{
			indexPrint(current->node, out, depth+1);
			
		}while(current = (muffetchild*)avlnext(current));
	}
}

void buildINDEX( int out){
	fprint(out, "\ninsert\nindex\n	BODY\n");
	muffetnode hkey = {.name="home"};
	muffetnode* home = (muffetnode*)avllookup(tree, &hkey,0);
	indexPrint(home, out, 1);

	fprint(out,"\n!\n");
}

void populateBREF(Avl * n, Biobuf *in, int out, int ){
	muffetnode * N = (muffetnode *) n;

	char * line;
	char* breflustr = "\nl\n%s\n\tBREF\n!\n";

	fprint(out, breflustr, N->name);
	line = Brdstr(in, '\n', 1);
	match[0].sp = match[0].ep = 0;
	if(regexec(valrp, line, match, MSZ)){
		N->bref = copymatch(1);
		free(line);
	} else { 
		while( line = Brdstr(in, '\n', 1)){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				free(line);
				break;
			}
			free(line);
		}
	}
}

void populateTYPE(Avl * n, Biobuf *in, int out, int ){
	muffetnode * N = (muffetnode *) n;

	char * line;
	char* typelustr = "\nl\n%s\n\tTYPE\n!\n";
	char * tmatch;

	fprint(out, typelustr, N->name);
	line = Brdstr(in, '\n', 1);
	match[0].sp = match[0].ep = 0;
	if(regexec(valrp, line, match, MSZ)){
		tmatch = copymatch(1);
		N->type = getmuffettype(tmatch);
		free(tmatch);
		free(line);
	} else { 
		while( line = Brdstr(in, '\n', 1)){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				free(line);
				break;
			}
			free(line);
		}
	}
}

#define REFSPERLINE 20
Avl * 
fillbodynode(char * line, muffet_type type, int bodycount){
	char * parts[REFSPERLINE+1];
	char * refs[REFSPERLINE];
	muffetbody * B = (muffetbody*) malloc(sizeof(muffetbody));
	B->line = line;
	B->index = bodycount;
	B->refs = nil;
	B->parts = nil;
	int refcount = 0;	

	if(type == o_LITERAL){
	//literal type disables meta processing
		B->refcount = refcount;
		return B;
	}
	parts[0] = line;
	
	for(int i = 0; line[i] != 0; ++i){
		if(line[i] == '{'){
			line[i] = 0;
			refs[refcount++] = line+i+1;
		}
		if(line[i] == '}'){
			line[i] = 0;
			parts[refcount] = line+i+1;
		}
	}
	B->refs  = (char**) malloc(sizeof(char*) *  refcount);
	B->parts = (char**) malloc(sizeof(char*) * (refcount + 1));
	B->refcount = refcount;
	for(int i = 0; i < refcount; ++i)
		B->refs[i] = refs[i];
	for(int i = 0; i <= refcount; ++i)
		B->parts[i] = parts[i];
	return B;
}

void
populateBODY(Avl * n, Biobuf *in, int out, int ){
	muffetnode * N = (muffetnode *) n;
	muffetbody * B = nil;

	char * line;
	char* bodylustr = "\nl\n%s\n\tBODY\n!\n";
	char * tmatch;
	int bodycount = 0;
	int firstline = 1;
	int nobody = 0;

	
	fprint(out, bodylustr, N->name);
	while (line = Brdstr(in, '\n', 1)){
	//print ("%s\n", line);
		if(firstline){
			firstline = 0;
			match[0].sp = match[0].ep = 0;
			if(regexec(namerp, line, match, MSZ)){
				//we found a host name, not BODY
				nobody = 1;
				free(line);
				continue; 
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(valrp, line, match, MSZ)){
				tmatch = copymatch(1);
				if(mcuistrcmp(tmatch,"BODY") == 0){
					free(tmatch);
					free(line);
					continue;
				} else { //single line body
					match[0].sp = match[0].ep = 0;
					if(regexec(valrp, line, match, MSZ)){
						tmatch = copymatch(1);
						B = (muffetbody*) fillbodynode(tmatch, N->type, ++bodycount);
						B = (muffetbody*) avlinsert(N->bodylines, B);
					}
					if(B)
						freemuffetbody(B);
					free(line);
					break;
				}
			}
		}
		if(nobody){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			free(line);
		} else {
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(listrp, line, match, MSZ)){
				tmatch = copymatch(1);
				B = (muffetbody*) fillbodynode(tmatch, N->type, ++bodycount);
				B = (muffetbody*) avlinsert(N->bodylines, B);
				if(B)
					freemuffetbody(B);
				free(line);
				continue; 
			}
			free(line);
		}
	}
}


void
populateLINK(Avl * n, Biobuf *in, int out, int ){
	muffetnode * N = (muffetnode *) n;
	muffetlink * L ;

	char * line;
	char* linklustr = "\nl\n%s\n\tLINK\n!\n";
	char* urllustr  = "\nl\n%s\n\t%s\n!\n";
	char* keylustr  = "\nl\n%s\n\tKEY\n!\n";
	// we added special keyname "key" which will return the names of all the keys
	// this way we can populate links in the form:
	/*********************************************************
	A LIST OF LINKS
		some site name : some site url
		another site name : another url
		etc. : etc.
	*********************************************************/
	//additionally we added a keyname 'list' 
	//this will get all the list names for lists which are of the form:
	/*********************************************************
	A LIST OF ITEMS
		something
		another thing
		and yet something else
	*********************************************************/
	//we just have to make sure we ignore the "well known" names (ie host, link, etc.)
	char * tmatch;
	int linkcount = 0;
	int firstline = 1;
	int nolink = 0;

	
	fprint(out, linklustr, N->name);
	while (line = Brdstr(in, '\n', 1)){
	//print ("%s\n", line);
		if(firstline){
			firstline = 0;
			match[0].sp = match[0].ep = 0;
			if(regexec(namerp, line, match, MSZ)){
				//we found a host name, not LINK
				nolink = 1;
				free(line);
				continue; 
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(valrp, line, match, MSZ)){
				tmatch = copymatch(1);
				if(mcuistrcmp(tmatch,"LINK") == 0){
					free(tmatch);
					free(line);
					continue;
				} else { //single line link
					match[0].sp = match[0].ep = 0;
					if(regexec(linkrp, line, match, MSZ)){
						L = (muffetlink*) malloc(sizeof(muffetlink));
						L->index = ++linkcount;
						L->name = copymatch(1);
						L->url = copymatch(2);
						L->desc = copymatch(3);
						L = (muffetlink*) avlinsert(N->links,L);
						if(L)
							freemuffetlink(L);
					}
					free(line);
					break;
				}
			}
		}
		if(nolink){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			free(line);
		} else {
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(linkrp, line, match, MSZ)){
				L = (muffetlink*) malloc(sizeof(muffetlink));
				L->index = ++linkcount;
				L->name = copymatch(1);
				L->url = copymatch(2);
				L->desc = copymatch(3);
				L = (muffetlink*) avlinsert(N->links,L);
				if(L)
					freemuffetlink(L);
			}
			free(line);
		}
	}
	fprint(out, keylustr, N->name);
	while (line = Brdstr(in, '\n', 1)){
		if (mcuistrcmp(line, "!") == 0 ){
			free(line);
			break;
		}
		if (mcuistrcmp(line, "	body") == 0 ||
			mcuistrcmp(line, "	link") == 0 ||
			mcuistrcmp(line, "	tags") == 0 ||
			mcuistrcmp(line, "	bref") == 0 ||
			mcuistrcmp(line, "	host") == 0 ||
			mcuistrcmp(line, "	date") == 0 ||
			mcuistrcmp(line, "	page") == 0 ||
			mcuistrcmp(line, "	pub") == 0 ||
			mcuistrcmp(line, "	type") == 0){
				free(line);
				continue;
		}
		match[0].sp = match[0].ep = 0;
		if(regexec(valrp, line, match, MSZ)){
			L = (muffetlink*) malloc(sizeof(muffetlink));
			L->index = ++linkcount;
			L->name = copymatch(1);
			//fprint(2, "%s-%s-\n",N->name, L->name);
			L->url = nil;
			L->desc = nil;
			L = (muffetlink*) avlinsert(N->links,L);
			if(L)
				freemuffetlink(L);
		}
	}

	L = (muffetlink*)avlmax(N->links);
	if(L){
		do{
			if(L && L->url == nil){
				fprint(out, urllustr, N->name, L->name);
				line = Brdstr(in, '\n', 1);
				match[0].sp = match[0].ep = 0;
				if(regexec(urlrp, line, match, MSZ)){
					L->url = copymatch(1);
					L->desc = copymatch(2);
				} /* else { //it's not a url - remove it from links
					if(L){
						L = (muffetlink*) avldelete(N->links, L);
						freemuffetlink(L);
						L = (muffetlink*) avlmax(N->links);
					}
				}*/
			}
		}while (L = (muffetlink*)avlprev(L));
	}
}

void
populateTAG(Avl * n, Biobuf *in, int out, int ){
	muffetnode * N = (muffetnode *) n;
	muffettag * T;

	char * line;
	char* taglustr = "\nl\n%s\n\tTAGS\n!\n";
	char * tmatch;
	int firstline = 1;
	int notag = 0;

	
	fprint(out, taglustr, N->name);
	while (line = Brdstr(in, '\n', 1)){
	//print ("%s\n", line);
		if(firstline){
			firstline = 0;
			match[0].sp = match[0].ep = 0;
			if(regexec(namerp, line, match, MSZ)){
				notag = 1;
				free(line);
				continue; 
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(valrp, line, match, MSZ)){
				tmatch = copymatch(1);
				if(mcuistrcmp(tmatch,"TAGS") == 0){
					free(tmatch);
					free(line);
					continue;
				} else { //single line link
					match[0].sp = match[0].ep = 0;
					if(regexec(valrp, line, match, MSZ)){
						T = (muffettag*) malloc(sizeof(muffettag));
						T->name = copymatch(1);
						T->node = (muffetnode*) avllookup(tree, T, 0);
						if(T->node == nil){
							fprint(2, "Tag node %s not found\n", T->name);
							freemuffettag(T);
							free(line);
							continue;
						}
						muffet_inc_ref * TaR = (muffet_inc_ref *)malloc(sizeof(muffet_inc_ref));
						TaR->name = N->name;
						TaR->node = N;
						TaR = (muffet_inc_ref*) avlinsert(T->node->inc_refs, TaR);
						//fprint(2, "put %s into %s \n",T->name,  T->node->name); 
						if(TaR)
							freemuffet_inc_ref(TaR);
						T = (muffettag*) avlinsert(N->tags, T);
						if(T)
							freemuffettag(T);
					}
					free(line);
					break;
				}
			}
		}
		if(notag){
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			free(line);
		} else {
			match[0].sp = match[0].ep = 0;
			if(regexec(eofrp, line, match, MSZ)){
				//the last line will be eof (!)
				free(line);
				break;
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(listrp, line, match, MSZ)){
				//fprint(2, "got here: %s \n", line);
				T = (muffettag*) malloc(sizeof(muffettag));
				T->name = copymatch(1);
				T->node = (muffetnode*) avllookup(tree, T, 0);
				if(T->node == nil){
					fprint(2, "Tag node %s not found\n", T->name);
					freemuffettag(T);
					free(line);
					continue;
				}
				muffet_inc_ref * TaR = (muffet_inc_ref *)malloc(sizeof(muffet_inc_ref));
				TaR->name = N->name;
				TaR->node = N;
				TaR = (muffet_inc_ref*) avlinsert(T->node->inc_refs, TaR);
				//fprint(2, "put %s into %s \n",N->name,  T->node->name); 
				if(TaR)
					freemuffet_inc_ref(TaR);
				T = (muffettag*) avlinsert(N->tags, T);
				if(T)
					freemuffetlink(T);
				
			}
			free(line);
		}
	}
}

void
calculateREFS(Avl * n, Biobuf *, int, int ){
	muffetnode * N = (muffetnode*) n;
	muffetnode * tN;
	muffetnode tNkey;
	char * tstr;
	int refcount = 0;
	muffet_out_ref * oR;

	muffet_inc_ref * iR;

	muffetbody* B = (muffetbody*) avlmin(N->bodylines);
	do{
		for(int i = 0 ; B && i < B->refcount; ++i){
			match[0].sp = match[0].ep = 0;
			if(regexec(erefrp, B->refs[i], match, MSZ)){
				tstr = copymatch(1);
				oR = (muffet_out_ref*) malloc(sizeof(muffet_out_ref));
				oR->index=++refcount;
				oR->name = copymatch(2);
				oR->type = getreftype(tstr);
				oR->desc = copymatch(3);
				sprint(B->refs[i], "%d", refcount);
				oR = (muffet_out_ref*) avlinsert(N->out_refs, oR);
				if(oR)
					freemuffet_out_ref(oR);
				continue;
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(urlrp,  B->refs[i], match, MSZ)){
				oR = (muffet_out_ref*) malloc(sizeof(muffet_out_ref));
				oR->index=++refcount;
				oR->name = copymatch(1);
				oR->type = getreftype("http");
				oR->desc = copymatch(2);
				sprint(B->refs[i], "%d", refcount);
				oR = (muffet_out_ref*) avlinsert(N->out_refs, oR);
				if(oR)
					freemuffet_out_ref(oR);
				continue;
			}
			match[0].sp = match[0].ep = 0;
			if(regexec(irefrp,  B->refs[i], match, MSZ)){
				oR = (muffet_out_ref*) malloc(sizeof(muffet_out_ref));
				oR->index=++refcount;
				oR->name = copymatch(1);
				oR->type = getreftype("intern");
				oR->desc = copymatch(2);
				sprint(B->refs[i], "%d", refcount);

				tNkey.name = oR->name;
				tN = (muffetnode*)avllookup(tree, &tNkey, 0);
				if(tN){
					iR = (muffet_inc_ref*) malloc(sizeof(muffet_inc_ref));
					iR->name = N->name;
					iR->node = N;
					iR = (muffet_inc_ref*) avlinsert(tN->inc_refs, iR);
					if(iR)
						freemuffet_inc_ref(iR);
				}
				
				oR = (muffet_out_ref*) avlinsert(N->out_refs, oR);
				if(oR)
					freemuffet_out_ref(oR);

				
			}

		}
	}while(B && (B = (muffetbody*) avlnext(B)));
}


void
main( int , char *[] )
{

	Biobuf * slovdat;
	int slovcmd;
	//char *line;

	irefrp = regcomp(irefrx);
	erefrp = regcomp(erefrx);
	valrp  = regcomp(valrx);
	eofrp  = regcomp(eofrx);
	namerp = regcomp(namerx);
	listrp = regcomp(listrx);
	linkrp = regcomp(linkrx);
	urlrp  = regcomp(urlrx);
	
	tree = avlcreate(mncmp);

	slovdat = Bopen("#s/slov.dat", OREAD);

	if(slovdat == nil){
		fprint(2, " slov.dat must be running! " );
		exits(nil);
	}
	
	slovcmd = open("#s/slov.cmd", OWRITE);

	if(slovcmd < 1){
		fprint(2, " slov.cmd must be running! " );
		exits(nil);
		
	}
	
	populateNAME(slovdat, slovcmd);
	
	walk(tree, slovdat, slovcmd, populateHOST, 1);
	buildINDEX(slovcmd);
	walk(tree, slovdat, slovcmd, populateBREF, 1);
	walk(tree, slovdat, slovcmd, populateTYPE, 1);
	walk(tree, slovdat, slovcmd, populateBODY, 1);
	walk(tree, slovdat, slovcmd, populateLINK, 1);
	walk(tree, slovdat, slovcmd, populateTAG, 1);
	walk(tree, slovdat, slovcmd, calculateREFS, 1); 
	walk(tree, nil, 1, printmuffetnode, 0);
	doGemini();
	doHTML();
	writeFeed(slovdat, slovcmd);


	free(irefrp);
	free(erefrp);
	free(valrp);
	free(eofrp);
	free(namerp);
	free(listrp);
	free(linkrp);
	free(urlrp);



}

